#!/bin/perl
#
# create.pl
#
# Gary Polhill 23 May 2001
#
# This is a script to create experiments for comparing two subpopulations
# against a third in a number of different environments. See the README file
# in this directory for instructions.

use strict;
use Cwd;
require "tools.pl";

# Get the list of environments we are comparing

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save messages in

open(LOG, ">create.log") || die "Cannot create log file create.log: $!\n";

# Check that the climate and economy toggle files are present, check they are
# consistent with the model file, and read in the model parameters -- we'll
# need them to check what to hold constant in the paired replicate tests.

my %model;
my %warnings;

&load_environments(\@envs, \%model);

# Now loop through the subpopulations, setting up the experiments. 

foreach my $dir (glob("*-v-*-x-*")) {
  if(!-d "$dir") {
    print LOG "WARNING: $dir matches *-v-*-x-*, but is not a directory -- it ".
      "will not be included in the experiment\n";
    next;
  }
  opendir(DIR, "$dir") || die "Could not open directory $dir";
  my @dirents = grep !/^\.\.?$/, readdir(DIR);
  closedir(DIR);
  if(scalar(@dirents) > 0) {
    print LOG "WARNING: $dir already has contents -- skipping to avoid any ".
      "over-writes\n";
    next;
  }
  $dir =~ /^(.*)-v-(.*)-x-(.*)$/;
  my $sp1 = $1;
  my $sp2 = $2;
  my $sp3 = $3;
  &check_subpops(\@envs, \%model, $sp2, $sp3);
  &check_subpops(\@envs, \%model, $sp1, $sp3);
  # Loop through all the environments
  for(my $env = 0; $env <= $#envs; $env++) {
    # Create the environment directory
    my $envdir = "$shortenvs[$env]";
    mkdir "$dir/$envdir", 0755 ||
      die "Could not create directory $dir/$envdir: $!\n";
    print LOG "NOTE: Created directory $dir/$envdir\n";
    # Check paired replicates -- everything is paired because the environment
    # is the same
    my($parcelPaired, $climatePaired, $economyPaired, $landUsePaired)
      = (1, 1, 1, 1);
    print LOG "NOTE: Pairing ";
    if($parcelPaired) {
      print LOG "parcels";
      print LOG (($climatePaired || $economyPaired || $landUsePaired) ?
		 ", " : "\n");
    }
    if($climatePaired) {
      print LOG "climate";
      print LOG (($economyPaired || $landUsePaired) ? ", " : "\n");
    }
    if($economyPaired) {
      print LOG "economy";
      print LOG ($landUsePaired ? ", " : "\n");
    }
    print LOG "land uses\n" if $landUsePaired;
    print LOG "nothing\n" if !($parcelPaired || $climatePaired
			       || $economyPaired || $landUsePaired);
    # Create the top-level subpopulation files
    my $spf1 = &create_subpop_file($dir, $envdir, [$sp1, $sp3]);
    my $spf2 = &create_subpop_file($dir, $envdir, [$sp2, $sp3]);
    # Create the report configuration file
    open(REPCFG, ">$dir/$envdir/Parcel.repcfg") ||
      die "Could not create configuration file $dir/$envdir/Parcel.repcfg: ".
	"$!\n";
    my $reportYear = $model{$envs[$env], 'maxYear'} - 1;
    print REPCFG "ParcelBitStringReport YearsToReport: 0\n"
      if $parcelPaired;
    print REPCFG "LandUseBitStringReport YearsToReport: 0\n"
      if $landUsePaired;
    print REPCFG "ClimateBitStringReport YearsToReport: Every 1\n"
      if $climatePaired;
    print REPCFG "EconomyBitStringReport YearsToReport: Every 1\n"
      if $economyPaired;
    print REPCFG "ParcelSubPopReport YearsToReport: $reportYear\n";
    print REPCFG "End\n";
    close(REPCFG);
    # Copy the subpopulation and environment support files
    foreach my $file ("${sp1}.sp", "${sp2}.sp", "${sp3}.sp", 
		      "${sp1}.ss", "${sp2}.ss", "${sp3}.ss",
		      $model{$envs[$env], 'climateToggleProbFile'},
		      $model{$envs[$env], 'economyToggleProbFile'}) {
      &cp($file, "$dir/$envdir/$file");
    }
    # Create the two model files -- one for each environment. Some model
    # parameters are written automatically, others are as per that in the
    # top level directory -- can't notify the user of this or they'll be
    # flooded with messages...
    foreach my $pairedmodelfile ([$spf1, "$sp1-v-$sp3-$envs[$env]"],
				 [$spf2, "$sp2-v-$sp3-$envs[$env]"]) {
      my($spf, $modelfile) = @$pairedmodelfile;
      open(MODEL, ">$dir/$envdir/$modelfile") || 
	die "Cannot create model file $dir/$envdir/$modelfile: $!\n";
      print MODEL '@begin', "\n";
      # First write all the automatic parameters
      foreach my $autoparam (['nSubPops', 2],
			     ['subPopFile', "$spf"],
			     ['infiniteTime', 0]) {
	my($param, $value) = @$autoparam;
	print MODEL "$param\t$value\n";
	if(defined($model{$envs[$env], $param})
	   && $model{$envs[$env], $param} ne $value) {
	  print LOG "WARNING: Over-riding value of $param in modelfile ".
	    "$envs[$env] ($model{$envs[$env], $param}) with value $value\n"
	      unless defined($warnings{$envs[$env], $param, 'override'});
	  $warnings{$envs[$env], $param, 'override'} = $value;
	}
      }
      # Write the paired replicate parameters. There will be a different
      # model file (created at run time) for each run in the experiment.
      # For now, anything in the model file to be replaced with the run
      # number is written %%RUN%%.
      if($parcelPaired) {
	print MODEL "landParcelFile\t${envdir}%%RUN%%.lp\n";
	print MODEL "useLandParcelFile\t1\n";
      }
      else {
	print MODEL "useLandParcelFile\t0\n";
      }
      if($climatePaired) {
	print MODEL "climateFile\t${envdir}%%RUN%%.clim\n";
	print MODEL "useClimateFile\t1\n";
      }
      else {
	print MODEL "useClimateFile\t0\n";
      }
      if($economyPaired) {
	print MODEL "economyFile\t${envdir}%%RUN%%.econ\n";
	print MODEL "useEconomyFile\t1\n";
      }
      else {
	print MODEL "useEconomyFile\t0\n";
      }
      if($landUsePaired) {
	print MODEL "landUseFile\t${envdir}%%RUN%%.lu\n";
	print MODEL "useLandUseFile\t1\n";
      }
      else {
	print MODEL "useLandUseFile\t0\n";
      }
      # Write out the rest of the parameters. Where these have not been
      # specified, either default values will be used, or an error message
      # will be generated.
      # Parameters with default values first
      foreach my $paramdefault (['neighbourhoodRadius', 1],
				['nInitXParcels', 1],
				['nInitYParcels', 1],
				['strategyChangeUnit', 0.0],
				['neighbourNoiseMax', 0.0],
				['neighbourNoiseMin', 0.0]) {
	my($param, $default) = @$paramdefault;
	if(defined($model{$envs[$env], $param})) {
	  print MODEL "$param\t$model{$envs[$env], $param}\n";
	}
	else {
	  print MODEL "$param\t$default\n";
	  print LOG "WARNING: $envs[$env] has no value for $param, using ".
	    "default value ($default)\n" 
	      unless defined($warnings{$envs[$env], $param, 'default'});
	  $warnings{$envs[$env], $param, 'default'} = $default;
	}
      }
      # Now the rest of the parameters
      foreach my $param ('environmentType',
			 'nLandUse',
			 'pLandUseDontCare',
			 'clumping',
			 'envXSize',
			 'envYSize',
			 'breakEvenThreshold',
			 'landParcelPrice',
			 'climateBSSize',
			 'economyBSSize',
			 'landParcelBSSize',
			 'climateToggleProbFile',
			 'economyToggleProbFile',
			 'maxYear'
			) {
	if(defined($model{$envs[$env], $param})) {
	  print MODEL "$param\t$model{$envs[$env], $param}\n";
	}
	else {
	  die "Cannot find parameter $param in model file $envs[$env]\n";
	}
      }
      # Finish off the file
      print MODEL '@end', "\n";
      close(MODEL);
    }
  }
}
close(LOG);

exit 0;
