# tools.pl
#
# Gary Polhill April 25, 2001
#
# This file contains a series of subroutines for all the various experiment
# setup programs to use.
#
# 0-5-3 Began version control on the experiment scripts. Updated scripts to
# cope with modifications to subpopulations for model0-5-3

# load_environments_uncomplainingly
# 
# Loads the environments, but doesn't do any checks

sub load_environments_uncomplainingly {
  my($envref, $modelref) = @_;

  &load_environments_temporamentally($envref, $modelref, 0);
}

# load_environments
#
# Takes a reference to an array of model files to load in, and a reference to
# an associative array that will contain all the parameter settings loaded in
# from the model file. Loads in all the model parameters, and checks
# consistency between the climate and economy as specified in the model file,
# and the values in the toggle probability files. Loads in the toggle
# probabilities.

sub load_environments {
  my($envref, $modelref) = @_;

  &load_environments_temporamentally($envref, $modelref, 1);
}

# load_environments_temporamentally
#
# OK, this was load_environments until the clumpy create.pl needed a
# load_environments that didn't complain about missing toggle probability files
# and what have you. So, the third argument is a flag, 1 if you want the
# complaints, 0 if you don't. 

sub load_environments_temporamentally {
  my($envref, $modelref, $complain) = @_;

  my $env;
  foreach $env (@$envref) {
    open(ENV, "<$env") || die "Cannot read model file $env: $!\n";
				# Complain about this regardless
    my($climSizeFlag, $econSizeFlag, $climToggFlag, $econToggFlag) = (0,0,0,0);
    my($climateSize, $economySize, $climateToggle, $economyToggle);
    while(my $line = <ENV>) {
      $line = &mychomp($line);
      # Ignore comments
      next if $line =~ /^\#/;
      # Ignore @begin and @end
      next if $line =~ /^\@/;
      # Get the Climate and Economy data
      if($line =~ /^climateBSSize\s+(\S+)$/) {
	$climateSize = $1;
	$climSizeFlag = 1;
      }
      elsif($line =~ /^economyBSSize\s+(\S+)$/) {
	$economySize = $1;
	$econSizeFlag = 1;
      }
      elsif($line =~ /^climateToggleProbFile\s+(\S+)$/) {
	$climateToggle = $1;
	$climToggFlag = 1;
      }
      elsif($line =~ /^economyToggleProbFile\s+(\S+)$/) {
	$economyToggle = $1;
	$econToggFlag = 1;
      }
      # Read in the contents of the model file
      if($line =~ /^(\S+)\s+(.+)$/) {
	$$modelref{$env, $1} = $2;
      }
    }
    close(ENV);
    if($complain) {
      # Check we got all the climate and economy data
      unless($climSizeFlag && $econSizeFlag && $climToggFlag
	     && $econToggFlag) {
	my $message = "";
	$message .= "Can't find climateBSSize in model file $env\n" 
	  if !$climSizeFlag;
	$message .= "Can't find economyBSSize in model file $env\n"
	  if !$econSizeFlag;
	$message .= "Can't find climateToggleProbFile in model file $env\n"
	  if !$climToggFlag;
	$message .= "Can't find economyToggleProbFile in model file $env\n"
	  if !$econToggFlag;
	die $message;
      }
      # Check the climate and economy files exist
      unless((-f "$climateToggle") && (-f "$economyToggle")) {
	my $message = "";
	$message .= "Can't find climate toggle file $climateToggle\n"
	  if !-f "$climateToggle";
	$message .= "Can't find economy toggle file $economyToggle\n"
	  if !-f "$economyToggle";
	die $message;
      }
    }
    # Check the climate and economy toggle files are consistent with the
    # model file and load in the toggle files, if present
    foreach my $check ([$climateToggle, $climateSize, 'climate'],
		       [$economyToggle, $economySize, 'economy']) {
      my($file, $size, $text) = @$check;
      my $element = 0;
      if(!open(FILE, "<$file")) {
	next if !$complain;
	die "Cannot open $text toggle file $file: $!\n";
      }
      while(my $line = <FILE>) {
	$line = &mychomp($line);
	if($line =~ /^NumberOfElements:\s+([0-9]+)$/) {
	  if($1 != $size) {
	    die "Number of elements in $file ($1) does not match $text".
	      " bit string size of $size in model file $env\n" if $complain;
	  }
	}
	elsif(($line =~ /^d\s+(.*)$/) && ($element < $size)) {
	  $$modelref{$env, $text.'ToggleArray', $element} = $1;
	  $element++;
	}
      }
      close(FILE);
      if($element < $size) {
	die "There are not enough lines ($element) in the $text toggle file".
	  " $file to specify all the toggle probabilities ($size)\n"
	    if $complain;
      }
    }
  }
}

# check_paired_replicates
#
# Takes a reference to the model associative array, and two environment keys.
# Returns Boolean values indicating whether the parameter settings for each
# environment mean that they can be paired for Land Parcels, Climate, Economy
# or Land Uses in a paired-replicate test.

sub check_paired_replicates {
  my($modelref, $env1, $env2) = @_;

  my($parcel, $climate, $economy, $landUse) = (0, 0, 0, 0);
  # We can pair land parcels if the clumping algorithm is the same and the
  # number of bits in the land parcel bit string is the same, and the
  # environments have the same size
  $parcel = 1 if(($$modelref{$env1, 'clumping'} eq
		  $$modelref{$env2, 'clumping'})
		 && ($$modelref{$env1, 'landParcelBSSize'} ==
		     $$modelref{$env2, 'landParcelBSSize'})
		 && ($$modelref{$env1, 'envXSize'} ==
		     $$modelref{$env2, 'envXSize'})
		 && ($$modelref{$env1, 'envYSize'} ==
		     $$modelref{$env2, 'envYSize'}));
  # We can pair the climate if the number of climate bits is the same, and
  # the two environments have the same climate toggle probabilities.
  $climate = 1 if(&same_toggle_probs($env1, $env2, 'climateToggleArray',
				     'climateBSSize', $modelref));
  # The economy can be paired for similar conditions to the climate.
  $economy = 1 if(&same_toggle_probs($env1, $env2, 'economyToggleArray',
				     'economyBSSize', $modelref));
  # We can pair land uses if the two environments have the same don't care
  # bit probabilities, and the same number of bits in the land parcel, climate
  # and economy. Yeah, and if we have the same number of land uses...
  $landUse = 1 if(($$modelref{$env1, 'pLandUseDontCare'} ==
		   $$modelref{$env2, 'pLandUseDontCare'})
		  && ($$modelref{$env1, 'landParcelBSSize'} ==
		      $$modelref{$env2, 'landParcelBSSize'})
		  && ($$modelref{$env1, 'climateBSSize'} ==
		      $$modelref{$env2, 'climateBSSize'})
		  && ($$modelref{$env1, 'economyBSSize'} ==
		      $$modelref{$env2, 'economyBSSize'})
		  && ($$modelref{$env1, 'nLandUse'} ==
		      $$modelref{$env2, 'nLandUse'})); 
  return($parcel, $climate, $economy, $landUse);
}

# create_subpop_file
#
# Takes a directory, a sub-directory and a reference to an array of 
# subpopulation files, and creates a top-level sub-population file.

sub create_subpop_file {
  my($dir, $envdir, $sparr) = @_;

  my @sps = @$sparr;
  my $directory;
  my $filename;
  if($envdir eq "NULL") {
    $directory = "$dir";
    $filename = $sps[0]."-only";
  }
  else {
    $directory = "$dir/$envdir";
    $filename = join('-v-', @sps);
  }
  open(SP, ">$directory/$filename.sp") ||
    die "Could not create subpopulation file ${dir}.sp in directory ".
      "$directory: $!\n";
  print SP "NumberOfSubPopulations: ", scalar(@sps), "\n";
  for(my $i = 0; $i <= $#sps; $i++) {
    print SP "SubPopulationFile ", $i + 1, ": $sps[$i].sp\n";
  }
  close(SP);
  return "$filename.sp";
}

# cp
#
# Copies a file from one location to another

sub cp {
  my($from, $to) = @_;

  open(CPI, "<$from") || die "Cannot read $from: $!\n";
  open(CPO, ">$to") || die "Cannot create $to: $!\n";
  print CPO <CPI>;
  close(CPO);
  close(CPI);
}

# same_toggle_probs
#
# Evaluates two toggle probability arrays to see if they are the same. The
# toggle probability arrays are stored in the model associative array. This
# subroutine needs keys for the two environments being compared, and keys
# for the toggle array parameter and the size of the toggle array, as well as
# a reference to the associative array itself.

sub same_toggle_probs {
  my($key1, $key2, $arrparam, $sizeparam, $modelref) = @_;

  if($$modelref{$key1, $sizeparam} != $$modelref{$key2, $sizeparam}) {
    return 0;
  }
  for(my $i = 0; $i < $$modelref{$key1, $sizeparam}; $i++) {
    if($$modelref{$key1, $arrparam, $i} != $$modelref{$key2, $arrparam, $i}) {
      return 0;
    }
  }
  return 1;
}

# check_subpops
#
# Given an array of sub-population files, puts a .sp and .ss suffix on each,
# and checks the files exist, and that they have reasonably sane contents.
# This latter bit requires that the model associative array is passed in,
# together with an array of model filenames.

sub check_subpops {
  my $envref = shift(@_);
  my $modelref = shift(@_);
  my $totalprob = 0;
  my $sptext = join(',', @_);
  foreach my $sp (@_) {
    die "Cannot find subpopulation file $sp.sp\n" if !-f "$sp.sp";
    die "Cannot find strategy selector file $sp.ss for subpopulation $sp\n"
      if !-f "$sp.ss";
    open(SP, "<$sp.sp") || die "Cannot read subpopulation file $sp.sp\n";
    my %sp;
    while(my $line = <SP>) {
      $line = &mychomp($line);
      next if $line =~ /^\#/;
      next if $line =~ /^\@/;
      if($line =~ /^(\S+)\s+(.+)$/) {
	$sp{$1} = $2;
      }
    }
    close(SP);
    # Check that the distribution for each parameter is either normal or unifm
    foreach my $dist ('habitImitateThresholdDist', 'imitateProbDist',
		      'neighbourWeightDist') {
      next if !defined($sp{$dist});
      if($sp{$dist} ne 'normal' && $sp{$dist} ne 'uniform') {
	die "$dist is not \"normal\" or \"uniform\" in subpop file $sp.sp\n";
      }
    }
    # Check the strategy selector file
    $totalprob += $sp{'probability'};
    if($sp{'strategySelectorFile'} ne "$sp.ss") {
      die "Strategy selector file in $sp.sp is not $sp.ss\n";
    }
    open(SS, "<$sp.ss") || die "Cannot read strategy selector file $sp.ss\n";
    my $lineno = 0;
    my $nstrategies;
    my $ns = 0;
    my(@above, @innov, @imit, @init);
    while(my $line = <SS>) {
      $line = &mychomp($line);
      $lineno++;
      if($lineno == 1) {
	if($line =~ /^NumberOfStrategyClasses:\s+([0-9]+)$/) {
	  $nstrategies = $1;
	}
	else {
	  die "Invalid format in $sp.ss, line 1: $line\n";
	}
      }
      elsif($lineno == 2) {
	if($line !~ /^Class\s+AboveThresholdProbability\s+BelowThresholdNonImitativeProbability\s+BelowThresholdImitativeProbability\s+InitialProbability$/) {
	  die "Invalid format in $sp.ss, line 2: $line\n";
	}
      }
      else {
	if($line =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/) {
	  $above[$ns] = [$1, $2];
	  $innov[$ns] = [$1, $3];
	  $imit[$ns] = [$1, $4];
	  $init[$ns] = [$1, $5];
	  $ns++;
	}
	elsif($line !~ /^$/) {
	  die "Invalid format in $sp.ss, line $lineno: $line\n";
	}
      }
    }
    close(SS);
    # Check that all the probabilities for each type sum to 1, and that if
    # NoStrategy is used, then the other parameters are set appropriately
    my %sum;
    $sum{'AboveThresholdProbability'} = 0;
    $sum{'BelowThresholdNonImitativeProbability'} = 0;
    $sum{'BelowThresholdImitativeProbability'} = 0;
    $sum{'InitialProbability'} = 0;
    for(my $i = 0; $i < $ns; $i++) {
      $sum{'AboveThresholdProbability'} += $above[$i][1];
      $sum{'BelowThresholdNonImitativeProbability'} += $innov[$i][1];
      $sum{'BelowThresholdImitativeProbability'} += $imit[$i][1];
      $sum{'InitialProbability'} += $init[$i][1];
      # If we've got a non-zero chance of selecting NoStrategy for the
      # innovative strategy, then either (a) the imitative probability must
      # be at least 1, or (b) the habit threshold must be zero. So we've an
      # error if non-zero NoStrategy probability and imitative probability not
      # equal to 1 and habit threshold more than zero.
      if(($innov[$i][0] eq 'NoStrategy' && $innov[$i][1] > 0)
	 && (($sp{'imitateProbDist'} eq 'uniform'
	      && ($sp{'imitateProbMin'} != 1.0 
		  || $sp{'imitateProbMax'} != 1.0))
	     || ($sp{'imitateProbDist'} eq 'normal'
		 && ($sp{'imitateProbMean'} != 1.0
		     || $sp{'imitateProbVar'} != 0.0)))
	 && (($sp{'habitImitateThresholdDist'} eq 'uniform'
	      && $sp{'habitImitateThresholdMax'} > 0.0)
	     || ($sp{'habitImitateThresholdDist'} eq 'normal'
		 && $sp{'habitImitateThresholdMean'} != 0.0
		 && $sp{'habitImitateThresholdVar'} != 0.0))) {
	die "NoStrategy used for non-imitative strategy with probability ".
	  "$innov[$i][1] in $sp.ss with imitative probability in $sp.sp set ".
	    "to a value not equal to 1 and possibility of non-zero habit ">
	      "threshold\n";
      }
      # If non-zero probability of imitative NoStrategy then there's an error
      # if the imitative probability is not zero and the habit threshold is
      # not zero.
      if(($imit[$i][0] eq 'NoStrategy' && $imit[$i][1] > 0)
	 && (($sp{'imitateProbDist'} eq 'uniform'
	      && ($sp{'imitateProbMin'} != 0.0 
		  || $sp{'imitateProbMax'} != 0.0))
	     || ($sp{'imitateProbDist'} eq 'normal'
		 && ($sp{'imitateProbMean'} != 0.0
		     || $sp{'imitateProbVar'} != 0.0)))
	 && (($sp{'habitImitateThresholdDist'} eq 'uniform'
	      && $sp{'habitImitateThresholdMax'} > 0.0)
	     || ($sp{'habitImitateThresholdDist'} eq 'normal'
		 && $sp{'habitImitateThresholdMean'} != 0.0
		 && $sp{'habitImitateThresholdVar'} != 0.0))) {
	die "NoStrategy used for imitative strategy with probability ".
	  "$imit[$i][1] in $sp.ss with non-zero imitative probabililty in ".
	    "$sp.sp and possibility of non-zero habit threshold\n";
      }
      # If there's a non-zero chance of initial NoStrategy then that's an error
      if($init[$i][0] eq 'NoStrategy' && $init[$i][1] > 0) {
	die "NoStrategy used for initial strategy with probability ".
	  "$init[$i][1] in $sp.ss\n";
      }
      # If there's a non-zero chance of above threshold NoStrategy then that's
      # an error if any environment has a breakeven threshold greater than
      # the habit threshold. For normally distributed habit thresholds, this
      # means zero variance.
      if($above[$i][0] eq 'NoStrategy' && $above[$i][1] > 0) {
	my $env;
	my @problemenvs;
	foreach $env (@$envref) {
	  if($sp{'habitImitateThresholdDist'} eq 'uniform') {
	    if(($$model{$env, 'breakEvenThreshold'}
		> $sp{'habitImitateThresholdMax'})
	       || ($$model{$env, 'breakEvenThreshold'}
		   > $sp{'habitImitateThresholdMin'})) {
	      $problemenvs[scalar(@problemenvs)] = $env;
	    }
	  }
	  elsif($sp{'habitImitateThresholdDist'} eq 'normal') {
	    if($sp{'habitImitateThresholdVar'} != 0.0
	       || ($$model{$env, 'breakEvenThreshold'}
		   > $sp{'habitImitateThresholdMean'})) {
	      $problemenvs[scalar(@problemenvs)] = $env;
	    }
	  }
	}
	if($#problemenvs >= 0) {
	  die "breakEvenThreshold is more than habit threshold in $sp.sp ".
	    "with NoStrategy used with non-zero probability in $sp.ss ".
	      "as the above-threshold strategy in the following model ".
		"files:\n".join("\n", @problemenvs)."\n";
	}
      }
    }
    my($key, $value);
    while(($key, $value) = each(%sum)) {
      if($value != 1.0) {
	die "$key does not sum to 1 ($value) in $sp.ss\n";
      }
    }
  }
  if($totalprob != 1.0) {
    die "Probabilities in sub-pop files $sptext don't sum to 1 ($totalprob)\n";
  }
}

# create_run_modelfile
#
# Takes the name of a model file, and a run number, and searches the modelfile
# for %%RUN%%. If it does not contain it, returns the name of the model file.
# If it does contain %%RUN%%, then replace each occurrence with the run number
# in an appropriately named new file, which is returned.

sub create_run_modelfile {
  my($origmodelfile, $run) = @_;

  my $cwd = getcwd();
  open(ORIGENV, "<$origmodelfile")
    || die "Cannot find $origmodelfile in directory $cwd: $!\n";
  my @lines = <ORIGENV>;
  close(ORIGENV);
  my $modelfile;
  if(grep(/%%RUN%%/, @lines)) {
    substr($modelfile = $origmodelfile, -6) = "-run$run.model";
    open(RUNENV, ">$modelfile")
      || die "Cannot create $modelfile in directory $cwd: $!\n";
    foreach my $line (@lines) {
      $line =~ s/%%RUN%%/$run/g;
      print RUNENV $line;
    }
    close(RUNENV);
  }
  else {
    $modelfile = $origmodelfile;
  }
  return($modelfile);
}

# no_use_bitstring_files
#
# Sets all the flags for useXXXFile parameters to 0

sub no_use_bitstring_files {
  my($modelfile) = @_;

  open(FILE, "<$modelfile")
    || die "Cannot find $modelfile in directory ".getcwd().": $!\n";
  my @lines = <FILE>;
  close(FILE);
  for(my $i = 0; $i <= $#lines; $i++) {
    if($lines[$i] =~ /^(\s*use[A-Za-z]+File\s+)1(\s*)$/) {
      $lines[$i] = $1."0".$2;
    }
  }
  open(FILE, ">$modelfile")
    || die "Cannot overwrite $modelfile in directory ".getcwd().": $!\n";
  foreach my $line (@lines) {
    print FILE $line;
  }
  close(FILE);
}

# datetime
#
# return the date in the format YYMMDDhhmmss

sub datetime {
  my ($s, $mi, $h, $d, $mo, $y) = localtime;

  return sprintf("%02d%02d%02d%02d%02d%02d", $y % 100, ++$mo, $d, $h, $mi, $s);
}

# get_short_envs
#
# Return a list of environments with the ".model" chopped off the back

sub get_short_envs {
  my @envs = @_;
  my @shortenvs;
  for(my $i = 0; $i <= $#envs; $i++) {
    # Chop the last six characters of each $env (which will be '.model') to get
    # a shorter name for the environments
    substr($shortenvs[$i] = $envs[$i], -6) = "";
  }
  return(@shortenvs);
}

# load_report_file
#
# Loads an entire report file in to the associative array reference passed as
# argument.

sub load_report_file {
  my($file, $report) = @_;

  open(REPORT, "<$file")
    || die "Cannot load report file ".getcwd()."/$report: $!\n";
  my $lineno = 0;
  while(my $line = <REPORT>) {
    $line = &mychomp($line);
    $lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'BEGIN') {
      if($cols[1] eq 'Parameters') {
	&load_report_parameters($file, REPORT, $report, \$lineno);
      }
      elsif($cols[1] eq 'Report for end of year:') {
	&load_report_reports($file, REPORT, $report, \$lineno, $cols[2]);
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $lineno):\n".
	  "$cols[0]\t$cols[1] <<< Don't know what to do with this\n";
      }
    }
  }
  close(REPORT);
}

# load_report_parameters
#
# Loads the parameters from the report file

sub load_report_parameters {
  my($file, $fd, $report, $lineno) = @_;
  
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      return if $cols[1] eq 'Parameters';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"Parameters\"\n";
    }
    $cols[0] =~ s/:$//;
    if($cols[0] eq 'Sub-population ID') {
      &load_report_subpop_parameters($file, $fd, $report, $lineno, $cols[1]);
    }
    elsif($cols[0] =~ /toggle probabilities$/) {
      my @togg = @cols[1 .. $#cols];
      $$report{'Parameter', $cols[0]} = [ @togg ];
    }
    elsif(scalar(@cols) > 2) {
      for(my $i = 1; $i <= $#cols; $i++) {
	$$report{'Parameter', $cols[0], $i - 1} = $cols[$i];
      }
    }
    else {
      $$report{'Parameter', $cols[0]} = $cols[1];
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tParameters\"\n";
}

# load_report_subpop_parameters
#
# Loads the subpopulation parameters from the report file, and also creates
# a strategy summary string. This can be used to compare two subpopulations
# to see if they have the same strategy quickly. It can also be used to 
# display the subpopulation.
	
sub load_report_subpop_parameters {
  my($file, $fd, $report, $lineno, $sp) = @_;

  my($paramflag, $strategyflag) = (0, 0);
  my($nabove, $nimit, $ninnov, $ninit) = (0, 0, 0, 0);
  my($lmc, $mstr, $hstr, $hsstr, $nstr, $pstr, $psstr, $npsstr, $isstr);
  #  |     |      |      |       |      |      |       |        Initial strategy str
  #  |     |      |      |       |      |      |       Innovative strategy string
  #  |     |      |      |       |      |      Imitative strategy string
  #  |     |      |      |       |      Imitative probability range string
  #  |     |      |      |       Neighbourhood weight range string
  #  |     |      |      Habit strategy string
  #  |     |      Habit threshold range string
  #  |     Memory range string
  #  Land manager class
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    $cols[0] =~ s/:$//;
    if($cols[0] eq 'End of parameters for sub-population ID') {
      my $summary = ($mstr eq "1" ? "" : "M$mstr")
	# ^^^ Put the memory string if it is not 1
	# Put the habit string if it is not 'NoStrategy' (remembering the
	# 'Strategy' will have been removed)
	.($hsstr eq 'No' ? "" : "H$hstr"."{$hsstr}")
	  # Put the neighbourhood weighting and the below threshold strategies
	  # so long as the habit threshold is not zero
	  .($hstr eq "0" ? "" :
	    # Put the neighbourhood weighting if not 1 and the imitative
	    # probability is not zero
	    (($nstr eq "1" || $pstr eq "0") ? "" : "N$nstr")
	    # For imitative probability of 1, only put the imitative strategy;
	    # for imitative probability of 0, only put the innovative strategy;
	    # otherwise put both.
	    .($pstr eq "1" ? "P{$psstr}" : ($pstr eq "0" ? "P-{$npsstr}" :
					    "P$pstr"."{$psstr}-{$npsstr}")))
	    # Put the initial strategy
	    ."I{$isstr}";
      $$report{'SubpopSummary', $sp} = $summary;
      return if $cols[1] eq $sp;
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"$sp\"\n";
    }
    elsif($cols[0]
	  eq 'Probability of land manager belonging to this sub-population') {
      $$report{'SubpopParam', $sp, 'Probability'} = $cols[1];
    }
    elsif($cols[0]
          eq 'Land Manager class') {
      $lmc = $cols[1];
    }
    elsif($cols[0] eq 'Strategy parameter ranges') {
      $paramflag = 1;
      $strategyflag = 0;
      next;
    }
    elsif($cols[0] eq 'Strategy method selector') {
      $paramflag = 0;
      $strategyflag = 1;
      next;
    }
    elsif($paramflag) {
      $$report{'SubpopParam', $sp, $cols[0], 'dist'} = $cols[1];
      if($cols[1] eq 'uniform') {
	$$report{'SubpopParam', $sp, $cols[0], 'min'} = $cols[2];
	$$report{'SubpopParam', $sp, $cols[0], 'max'} = $cols[3];
      }
      else {
	$$report{'SubpopParam', $sp, $cols[0], 'mean'} = $cols[2];
	$$report{'SubpopParam', $sp, $cols[0], 'var'} = $cols[3];
      }
      my $str;
      if($cols[1] eq 'uniform') {
        $str = (($cols[2] == $cols[3]) ? "$cols[2]" : "[$cols[2],$cols[3]]");
      }
      elsif($cols[1] eq 'normal') {
	$str = (($cols[3] == 0) ? "$cols[2]" : "N~($cols[2],$cols[3])");
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0]\t$cols[1] <<< Should say \"uniform\" or \"normal\"\n";
      }
      $mstr = $str if $cols[0] eq 'Memory size';
      $hstr = $str if $cols[0] eq 'Habit-imitation yield threshold';
      $pstr = $str if $cols[0] eq 'Probability of using imitative strategy '.
	'when yield below habit-imitation threshold';
      $nstr = $str if $cols[0] eq 'Neighbourhood weighting';
    }
    elsif($strategyflag) {
      next if $cols[0] eq 'Class';
      # Create a short name for the strategy by losing the Strategy at the end:
      # e.g. MatchWeightedOptimumStrategy -> MatchWeightedOptimum
      (my $short = $cols[0]) =~ s/Strategy$//;
      if($cols[1] > 0) {
	$$report{'SubpopParam', $sp, 'Above', $nabove} = [$cols[0], $cols[1]];
	$hsstr .= ($cols[1] == 1 ? "$short" : "[$short,$cols[1]]");
	$nabove++;
      }
      if($cols[2] > 0) {
	$$report{'SubpopParam', $sp, 'Innov', $ninnov} = [$cols[0], $cols[2]];
	$ninnov++;
	$npsstr .= ($cols[2] == 1 ? "$short" : "[$short,$cols[2]]");
      }
      if($cols[3] > 0) {
	$$report{'SubpopParam', $sp, 'Imit', $nimit} = [$cols[0], $cols[3]];
	$nimit++;
	$psstr .= ($cols[3] == 1 ? "$short" : "[$short,$cols[3]]");
      }
      if($cols[4] > 0) {
	$$report{'SubpopParam', $sp, 'Init', $ninit} = [$cols[0], $cols[4]];
	$ninit++;
	$isstr .= ($cols[4] == 1 ? "$short" : "[$short,$cols[4]]");
      }
    }
    else {
      # I *think* this is the only error you'll get -- once the paramflag is
      # set, it'll just keep reading...
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"Strategy parameter ranges\"\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for ".
      "\"End of parameters for sub-population ID:\t$sp\"\n";
}

# load_report_reports
#
# Loads in a report for a particular year

sub load_report_reports {
  my($file, $fd, $report, $lineno, $year) = @_;

  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'BEGIN') {
      if($cols[1] eq 'ParcelSubPopReport') {
	&load_report_ParcelSubPop($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'ClumpinessReport') {
	&load_report_Clumpiness($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'ParcelBitStringReport') {
	&load_report_ParcelBitString($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'ClimateBitStringReport') {
	&load_report_ClimateBitString($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'EconomyBitStringReport') {
	&load_report_EconomyBitString($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'LandUseBitStringReport') {
	&load_report_LandUseBitString($file, $fd, $report, $lineno, $year);
      }
      elsif($cols[1] eq 'MeanSubPopParamParcelReport') {
	&load_report_MeanSubPopParamParcel($file, $fd, $report, $lineno,$year);
      }
      else {
	&load_report_generic($file, $fd, $report, $lineno, $year, $cols[1]);
      }
      push(@{$$report{$cols[1], 'ReportingYears'}}, $year);
    }
    elsif($cols[0] eq 'END') {
      return if($cols[1] eq 'Report for end of year:' && $cols[2] == $year);
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"Report for end of year:\"\n"
	  if($cols[1] ne 'Report for end of year:');
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1]\t$cols[2] <<< Should say \"$year\"\n";
    }
    else {
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"BEGIN\" or \"END\"\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tReport for end of year:\t$year\"\n";
}

# load_report_generic
#
# Generic report loader. This just creates an array of references to the array
# of columns in each line, one per line, so a report like this:
#
# BEGIN	ParcelSubPopReport
# Sub-population ID:	1	Land parcels:	56
# Sub-population ID:	2	Land parcels:	44
# END	ParcelSubPopReport
#
# Would be entered in the associative array like this:
#
# $$report{'ParcelSubPopReport', $year} = 
#   [ ['Sub-population ID:', 1, 'Land parcels:', 56],
#     ['Sub-population ID:', 2, 'Land parcels:', 44] ];

sub load_report_generic {
  my($file, $fd, $report, $lineno, $year, $type) = @_;

  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      return if $cols[1] eq $type;
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"$type\"\n";
    }
    push(@{$$report{$type, $year}}, [ @cols ]);
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\t$type\"\n";
}

# load_report_ParcelSubPop
#
# Report loader for the ParcelSubPopReport. This is an array of hashes of
# subpop ID to number of land parcels, so a report like this:
#
# BEGIN	ParcelSubPopReport
# Sub-population ID:	1	Land parcels:	56
# Sub-population ID:	2	Land parcels:	44
# END	ParcelSubPopReport
#
# Would be entered in the associative array like this:
#
# $$report{'ParcelSubPopReport', $year} = 
#   { 1 => 56,
#     2 => 44 };

sub load_report_ParcelSubPop {
  my($file, $fd, $report, $lineno, $year) = @_;

  my %parcel;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      $$report{'ParcelSubPopReport', $year} = \%parcel;
      return if $cols[1] eq 'ParcelSubPopReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"ParcelSubPopReport\"\n";
    }
    elsif($cols[0] eq 'Sub-population ID:' && $cols[2] eq 'Land parcels:') {
      $parcel{$cols[1]} = $cols[3];
    }
    else {
      if($cols[0] eq 'Sub-population ID:') {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0]\t$cols[1]\t$cols[2] <<< Should say \"Land parcels:\"\n";
      }
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"Sub-population ID:\" or \"END\"\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tParcelSubPopReport\"\n";
}

# load_report_Clumpiness
#
# Report loader for the ClumpinessReport. This creates a hash to put the
# histogram in (if present), and always has a hash for the median, and
# normalised median. So, a report like this:
#
# BEGIN ClumpinessReport
# Number of matching bits	Frequency	Frequency as proportion
# 0	0	0
# 1	0	0
# 2	1	0.005
# 3	0	0
# 4	4	0.02
# 5	12	0.06
# 6	15	0.075
# 7	15	0.075
# 8	12	0.06
# 9	25	0.125
# 10	32	0.16
# 11	6	0.03
# 12	17	0.085
# 13	32	0.16
# 14	18	0.09
# 15	2	0.01
# 16	9	0.045
# Median number of matching bits	10
# Median proportion of matching bits	0.625
# Mean number of matching bits  <number I can't be bothered to calculate>
# Mean proportion of matching bits      <ditto>
# END ClumpinessReport
#
# Would go into the report hash like this:
#
# $$report{'ClumpinessReport', $year} = {
#	'HistogramPresent' => 1,
#	'Histogram' => {	0 => 0,		1 => 0,		2 => 0.005,
#				3 => 0,		4 => 0.02,	5 => 0.06,
#				6 => 0.075,	7 => 0.075,	8 => 0.06, ...
#				14 => 0.09,	15 => 0.01,	16 => 0.045 },
#	'Median' => 10,
#	'NormalisedMedian' => 0.625,
#       'Mean' => ???,
#       'NormalisedMean' => ??? };

sub load_report_Clumpiness {
  my($file, $fd, $report, $lineno, $year) = @_;

  $$report{'ClumpinessReport', $year}{HistogramPresent} = 0;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      return if $cols[1] eq 'ClumpinessReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"ClumpinessReport\"\n";
    }
    elsif($cols[0] eq 'Number of matching bits'
	  && $cols[1] eq 'Frequency'
	  && $cols[2] eq 'Frequency as proportion') {
      $$report{'ClumpinessReport', $year}{HistogramPresent} = 1;
    }
    elsif($cols[0] =~ /^[0-9]+$/
	  && $$report{'ClumpinessReport', $year}{HistogramPresent} == 1) {
      $$report{'ClumpinessReport', $year}{Histogram}{$cols[0]} = $cols[2];
				# Ignore the non-normalised histogram values
    }
    elsif($cols[0] eq 'Median number of matching bits') {
      $$report{'ClumpinessReport', $year}{Median} = $cols[1];
    }
    elsif($cols[0] eq 'Median proportion of matching bits') {
      $$report{'ClumpinessReport', $year}{NormalisedMedian} = $cols[1];
    }
    elsif($cols[0] eq 'Mean number of matching bits') {
      $$report{'ClumpinessReport', $year}{Mean} = $cols[1];
    }
    elsif($cols[0] eq 'Mean proportion of matching bits') {
      $$report{'ClumpinessReport', $year}{NormalisedMean} = $cols[1];
    }
    else {
      if($cols[0] eq 'Number of matching bits') {
	if($cols[1] eq 'Frequency') {
	  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	    "$cols[0]\t$cols[1]\t$cols[2] <<< Should say \"Frequency as ".
	      "proportion\"\n";
	}
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0]\t$cols[1] <<< Should say \"Frequency\"\n";
      }
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"Median/Mean number of matching bits\" or ".
	  "\"Median/Mean proportion of matching bits\"".
	    ($$report{'ClumpinessReport', $year}{HistogramPresent} == 1 ?
	     " or <a number of land parcel bits for the histogram>\n" :
	     " or \"Number of matching bits\"\n");
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tClumpinessReport\"\n";
}

# load_report_ParcelBitString
#
# Report loader for the ParcelBitStringReport. This creates an array of arrays
# of the bit strings, so to access the bitstring of a land parcel, use
#
# ${@{$$report{'ParcelBitStringReport', $year}}}[$x][$y]

sub load_report_ParcelBitString {
  my($file, $fd, $report, $lineno, $year) = @_;

  my @parcel;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      $$report{'ParcelBitStringReport', $year} = [ @parcel ];
      return if $cols[1] eq 'ParcelBitStringReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"ParcelBitStringReport\"\n";
    }
    elsif($cols[0] eq 'Land parcel') {
      if($cols[1] eq 'X:') {
	if($cols[3] eq 'Y:') {
	  if($cols[5] eq 'Bitstring:') {
	    $parcel[$cols[2]][$cols[4]] = $cols[6];
	  }
	  else {
	    die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	      join('\t', @cols[0..5])." <<< Should say \"Bitstring:\"\n";
	  }
	}
	else {
	  die"Error in report file ".getcwd()."/$file (line $$lineno):\n".
	    join('\t', @cols[0..3])." <<< Should say \"Y:\"\n";
	}
      }
      else {
	die"Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  join('\t', @cols[0..1])." <<< Should say \"X:\"\n";
      }
    }
    else {
      die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"Land parcel\" or \"END\"\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tParcelBitStringReport\"\n";
}

# load_report_LandUseBitString
#
# Report loader for the land use bit strings. This creates an array of hashes,
# so a report like this:
#
# BEGIN	LandUseBitStringReport
# Land use	Match bitstring:	0011	Don't care bitstring:	0000
# Land use	Match bitstring:	1111	Don't care bitstring:	0000
# Land use	Match bitstring:	0001	Don't care bitstring:	0000
# END	LandUseBitStringReport
#
# ... would get converted to:
#
# $$report{'LandUseBitStringReport', $year} =
#   [ { 'match'    => '0011',
#       'dontcare' => '0000' },
#     { 'match'    => '1111',
#       'dontcare' => '0000' },
#     { 'match'    => '0001',
#       'dontcare' => '0000' } ];

sub load_report_LandUseBitString {
  my($file, $fd, $report, $lineno, $year) = @_;

  my @uses;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      $$report{'LandUseBitStringReport', $year} = [ @uses ];
      return if $cols[1] eq 'LandUseBitStringReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"LandUseBitStringReport\"\n";
    }
    elsif($cols[0] eq 'Land use') {
      if($cols[1] eq 'Match bitstring:') {
	if($cols[3] eq "Don't care bitstring:") {
	  push(@uses, { 'match' => $cols[2], 'dontcare' => $cols[4] });
	}
	else {
	  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	      join('\t', @cols[0..3]).
		" <<< Should say \"Don't care bitstring:\"\n";
	}
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  join('\t', @cols[0..1])." <<< Should say \"Match bitstring:\"\n";
      }
    }
    else {
      die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say \"Land use\" or \"END\"\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tLandUseBitStringReport\"\n";
}

# load_report_ClimateBitString
#
# Loads in the report containing the climate bit string.

sub load_report_ClimateBitString {
  my($file, $fd, $report, $lineno, $year) = @_;

  my $flag = 1;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      if($flag) {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Should say \"Climate\" (found end of report before ".
	    "reading in the bitstring!)\n";
      }
      return if $cols[1] eq 'ClimateBitStringReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"ClimateBitStringReport\"\n";
    }
    elsif($cols[0] eq 'Climate') {
      if($flag) {
	if($cols[1] eq 'Bitstring:') {
	  $$report{'ClimateBitStringReport', $year} = $cols[2];
	}
	else {
	  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	    "$cols[0]\t$cols[1] <<< Should say \"Bitstring:\"\n";
	}
	$flag = 0;
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Should say \"END\" (not expecting 2 climates for the".
	    "same year)\n";
      }
    }
    else {
      die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say ".($flag ? "\"Climate\"" : "\"END\"")."\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tClimateBitStringReport\"\n";
}

# load_report_EconomyBitString
#
# Loads in the report containing the economy bit string.

sub load_report_EconomyBitString {
  my($file, $fd, $report, $lineno, $year) = @_;

  my $flag = 1;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      if($flag) {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Should say \"Economy\" (found end of report before ".
	    "reading in the bitstring!)\n";
      }
      return if $cols[1] eq 'EconomyBitStringReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"EconomyBitStringReport\"\n";
    }
    elsif($cols[0] eq 'Economy') {
      if($flag) {
	if($cols[1] eq 'Bitstring:') {
	  $$report{'EconomyBitStringReport', $year} = $cols[2];
	}
	else {
	  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	    "$cols[0]\t$cols[1] <<< Should say \"Bitstring:\"\n";
	}
	$flag = 0;
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Should say \"END\" (not expecting 2 economies for the".
	    "same year)\n";
      }
    }
    else {
      die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say ".($flag ? "\"Economy\"" : "\"END\"")."\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tEconomyBitStringReport\"\n";
}

# load_report_MeanSubPopParamParcel
#
# A routine to load in the MeanSubPopParamParcelReport. This will convert
# a report like this:
#
# BEGIN MeanSubPopParamParcelReport
# Mean neighbourhood weight (over LPs):  0.79
# Mean imitation probability (over LPs): 0.64
# Mean habit threshold (over LPs):       14.36
# Mean memory size (over LPs):   4.28
# END   MeanSubPopParamParcelReport
#
# ... into:
#
# $$report{'MeanSubPopParamParcelReport', $year} = 
#   { 'neighbourhood weight'  => 0.79,
#     'imitation probability' => 0.64,
#     'habit threshold'       => 14.36,
#     'memory size'           => 4.28 };

sub load_report_MeanSubPopParamParcel {
  my($file, $fd, $report, $lineno, $year) = @_;

  my %flag = { 'neighbourhood weight' => 1,
	       'imitation probability' => 1,
	       'habit threshold' => 1,
	       'memory size' => 1 };
  my %param;
  while(my $line = <$fd>) {
    $line = &mychomp($line);
    $$lineno++;
    my @cols = split(/\t/, $line);
    if($cols[0] eq 'END') {
      if(join('', values(%flag)) ne '0000') {
	my @missing;
	while(my($key, $value) = each(%flag)) {
	  push(@missing, $key) if($value != 0);
	}
	die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	  "Missing ".join(',', @missing)." values in ".
	    "MeanSubPopParamParcelReport\n";
      }
      $$report{'MeanSubPopParamParcelReport', $year} = \%param;
      return if $cols[1] eq 'MeanSubPopParamParcelReport';
      die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	"$cols[0]\t$cols[1] <<< Should say \"MeanSubPopParamParcelReport\"\n";
    }
    elsif($cols[0] eq 'Mean neighbourhood weight (over LPs):') {
      if($flag{'neighbourhood weight'}) {
	$param{'neighbourhood weight'} = $cols[1];
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Not expecting this second neighbourhood weight line\n";
      }
      $flag{'neighbourhood weight'} = 0;
    }
    elsif($cols[0] eq 'Mean imitation probability (over LPs):') {
      if($flag{'imitation probability'}) {
	$param{'imitation probability'} = $cols[1];
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Not expecting this second imitation probability ".
	    "line\n";
      }
      $flag{'imitation probability'} = 0;
    }
    elsif($cols[0] eq 'Mean habit threshold (over LPs):') {
      if($flag{'habit threshold'}) {
	$param{'habit threshold'} = $cols[1];
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Not expecting this second habit threshold line\n";
      }
      $flag{'habit threshold'} = 0;
    }
    elsif($cols[0] eq 'Mean memory size (over LPs):') {
      if($flag{'memory size'}) {
	$param{'memory size'} = $cols[1];
      }
      else {
	die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
	  "$cols[0] <<< Not expecting this second memory size line\n";
      }
      $flag{'memory size'} = 0;
    }
    else {
      die "Error in report file".getcwd()."/$file (line $$lineno):\n".
	"$cols[0] <<< Should say ".
	  ($flag{'neighbourhood weight'} ? "\"Mean neighbourhood weight...\""
	   : ($flag{'imitation probability'} 
	      ? "\"Mean imitation probability...\""
	      : ($flag{'habit threshold'} ? "\"Mean habit threshold...\""
		 : ($flag{'memory size'} ? "\"Mean memory size...\""
		    : "\"END\""))))."\n";
    }
  }
  die "Error in report file ".getcwd()."/$file (line $$lineno):\n".
    "Found EOF while looking for \"END\tMeanSubPopParamParcelReport\"\n";
}

# compare_parcel_bitstrings
#
# Check the land parcel bitstrings match between two runs loaded in from their
# report files, passed as references associative arrays. Return 1 if they match
# and 0 if they don't. The match is complete if each report has bitstrings for
# the same years, and in each year, that bitstring is the same for a land
# parcel of the corresponding co-ordinates.

sub compare_parcel_bitstrings {
  my($r1, $r2, $msg) = @_;

  # First check the years
  my @y1 = @{$$r1{'ParcelBitStringReport', 'ReportingYears'}};
  my @y2 = @{$$r2{'ParcelBitStringReport', 'ReportingYears'}};
  if($#y1 != $#y2) {
    $$msg = "Different number of years reported on for parcel bit string ".
      "(".scalar(@y1).", ".scalar(@y2).")";
    return 0;
  }
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    if($y1[$yr] != $y2[$yr]) {
      $$msg = "Different year reported on in year ".($yr + 1).
	" ($y1[$yr], $y2[$yr])";
      return 0;
    }
  }
  # Check the environment sizes are the same. Different environment sizes
  # mean different land parcel bit strings -- and probably that the sim failed
  # in any case, since it won't let you load in land parcels for a differently
  # sized environment...
  my $xs1 = $$r1{'Parameter', 'Number of horizontal cells in the Environment'};
  my $ys1 = $$r1{'Parameter', 'Number of vertical cells in the Environment'};
  my $xs2 = $$r2{'Parameter', 'Number of horizontal cells in the Environment'};
  my $ys2 = $$r2{'Parameter', 'Number of vertical cells in the Environment'};
  if($xs1 != $xs2 || $ys1 != $ys2) {
    $$msg = "Different environment sizes ($xs1 x $ys1, $xs2 x $ys2)";
    return 0;
  }
  # Now check the bit strings
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    my(@p1, @p2);
    @p1 = @{$$r1{'ParcelBitStringReport', $y1[$yr]}};
    @p2 = @{$$r2{'ParcelBitStringReport', $y2[$yr]}};
    for(my $x = 0; $x < $xs1; $x++) {
      for(my $y = 0; $y < $ys1; $y++) {
	if($p1[$x][$y] ne $p2[$x][$y]) {
	  $$msg = "Different bit strings for land parcel at ($x, $y) in year ".
	    "$y1[$y2] ($p1[$x][$y], $p2[$x][$y])";
	  return 0;
	}
      }
    }
  }
  return 1;
}

# compare_landuse_bitstrings
#
# Check the bitstrings of the land uses in two runs. As for the corresponding
# parcel subroutine, references to report associative arrays are passed in
# after loading from the appropriate file. The match is complete if the same
# years are reported on and the bitstrings are the same in each year

sub compare_landuse_bitstrings {
  my($r1, $r2) = @_;

  # First check the years
  my @y1 = @{$$r1{'ParcelBitStringReport', 'ReportingYears'}};
  my @y2 = @{$$r2{'ParcelBitStringReport', 'ReportingYears'}};
  return 0 if($#y1 != $#y2);
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    return 0 if($y1[$yr] != $y2[$yr]);
  }
  # Now check the bitstrings
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    my @lu1 = @{$$r1{'LandUseBitStringReport', $y1[$yr]}};
    my @lu2 = @{$$r2{'LandUseBitStringReport', $y2[$yr]}};
    return 0 if($#lu1 != $#lu2);
    for(my $i = 0; $i <= $#lu1; $i++) {
      my %bs1 = %{$lu1[$i]};
      my %bs2 = %{$lu2[$i]};
      return 0 if($bs1{'match'} ne $bs2{'match'}
		  || $bs1{'dontcare'} ne $bs2{'dontcare'});
    }
  }
  return 1;
}

# compare_climate_economy_bitstrings
#
# Check for the same bitstrings in the report passed as argument. Will be used
# by climate and economy comparison routines

sub compare_climate_economy_bitstrings {
  my($r1, $r2, $report) = @_;

  # First check the years
  my @y1 = @{$$r1{$report, 'ReportingYears'}};
  my @y2 = @{$$r2{$report, 'ReportingYears'}};
  return 0 if($#y1 != $#y2);
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    return 0 if($y1[$yr] != $y2[$yr]);
  }
  # Now check the bitstrings
  for(my $yr = 0; $yr <= $#y1; $yr++) {
    return 0 if $$r1{$report, $yr} ne $$r2{$report, $yr};
  }
  return 1;
}

# compare_climate_bitstrings
#
# Check the bitstrings of the climates match in two runs. Since this is the
# same for the economy, call a metasubroutine to do this.

sub compare_climate_bitstrings {
  my($r1, $r2) = @_;

  return &compare_climate_economy_bitstrings($r1, $r2, 
					     'ClimateBitStringReport');
}

# compare_economy_bitstrings
#
# As for climate, but the economy

sub compare_economy_bitstrings {
  my($r1, $r2) = @_;

  return &compare_climate_economy_bitstrings($r1, $r2,
					     'EconomyBitStringReport');
}

# combinations
#
# Return the number of combinations of n things taken r at a time:
# n!/((n-r)!r!)

sub combinations {
  my($n, $r) = @_;

  my $comb = 1;
  for(my $i = $n; $i > ($n - $r); $i--) {
    $comb *= $i;
  }
  for(my $i = $r; $i > 1; $i--) {
    $comb /= $i;
  }
  return $comb;
}

# onetailbinomial
#
# Return the probability that X <= $wins in the specified number of trials if
# the probability of a win is 0.5 

sub onetailbinomial {
  my($trials, $wins) = @_;

  my($pvalue) = 0.0;
  for(my $i = 0; $i <= $wins; $i++) {
    $pvalue +=
      &combinations($trials, $i) * (0.5 ** $i) * (0.5 ** ($trials - $i));
  }
  return $pvalue;
}

# testgtbinomial
#
# Return a boolean indicating whether or not the null hypothesis is rejected
# at the given level of significance for a test that p > 0.5.

sub testgtbinomial {
  my($trials, $wins, $siglevel, $pvalue) = @_;

  if(($$pvalue = &onetailbinomial($trials, $trials - $wins)) < $siglevel) {
    return 1;
  }
  else {
    return 0;
  }
}
  
# testltbinomial
#
# Return a boolean indicating whether or not the null hypothesis is rejected
# at the given level of significance for a test that p < 0.5.

sub testltbinomial {
  my($trials, $wins, $siglevel, $pvalue) = @_;

  if(($$pvalue = &onetailbinomial($trials, $wins)) < $siglevel) {
    return 1;
  }
  else {
    return 0;
  }
}

# testnebinomial
#
# Return a boolean indicating whether or not the null hypothesis is rejected
# at the given level of significance for a test that p != 0.5. (This is a
# two-tailed test.)

sub testnebinomial {
  my($trials, $wins, $siglevel, $pvalue) = @_;

  if(($$pvalue = &onetailbinomial($trials, $trials - $wins))
     < ($siglevel / 2)) {
    return 1;
  }
  elsif(($$pvalue = &onetailbinomial($trials, $wins)) < ($siglevel / 2)) {
    return 1;
  }
  else {
    return 0;
  }
}

# testkruskal_wallis
#
# Return a boolean indicating whether or not the null hypothesis that each of
# a number of different samples have been taken from the same distribution
# should be rejected. The data structure to pass in is a reference to an array
# of samples, where each sample is an array of observed values. This subroutine
# uses the Chi-Squared approximation, which is OK for "moderate to large" 
# numbers of samples, but for low numbers of samples (and certainly two or
# three), there is a strong probability of type II error (but never type I) if
# the Chi-Squared approximation is used.

sub testkruskal_wallis {
  my($data, $siglevel, $pvalue, $Tstat) = @_;

  my(%ranks, @flat, @samples, $df);

  @samples = &rankdata($data, \%ranks, \@flat);
  ($$Tstat, $df) = &calckwstat($data, \%ranks);

  return(&testchisq($$Tstat, $df, $siglevel, $pvalue));
}

# testMCkruskal_wallis
#
# Monte Carlo estimation should be used for small sample sizes. This is done
# by computing the Kruskall-Wallis test statistic for a large number of
# different randomly generated permutations of the observations. You then
# compare the Test statistic for the sample size passed in with how many of
# the randomly generated permutations had higher and lower test statistics.
#
# The number of permutations to do is passed in with the accuracy parameter.
# Ten to the power accuracy random permutations will be generated. This means
# the p-value can be calculated as a count of the number of permutations with
# greater Test statistic than the original data, right-shifted by the accuracy.
# The limitation on accuracy will therefore be the point at which this count
# over-flows. However, the likelihood is that this will occur after the count
# of the number of random permutations overflows, so this will also be needed.

sub testMCkruskal_wallis {
  my($data, $accuracy, $siglevel, $pvalue, $Tstat) = @_;

  my(%ranks, @flat, @samples, $df);

  @samples = &rankdata($data, \%ranks, \@flat);
  ($$Tstat, $df) = &calckwstat($data, \%ranks);

  my($tot, $ovtot) = (0, 0);	# Total and overflow detector
  my $ovtotflag = 0;
  my($ngt, $ovngt, $ngtovtot) = (0, 0, 0);
				# Counter of the number of permutations with
				# greater Kruskal-Wallis test statistic than
				# the original data, together with overflow
				# detector, and count value when the total
				# overflows
  my $allgtflag = 1;		# There's a little exception to take care of,
				# which is when all the permutations are
				# greater than the sample. Then the pvalue = 1.
  my(@acccount);		# Counters for the accuracy
  for(my $i = 0; $i <= $accuracy; $i++) {
				# Use one extra counter to act as the test for
				# when we've done all we were supposed to
    $acccount[$i] = 10;
  }
  while($acccount[0] == 10) {
    # Increment the total number of permutations done, and check for overflows
    if(!$ovtotflag) {
      $tot++;
      if($tot <= $ovtot) {	# The increment has overflowed $tot -- remember
				# $ovtot and $ngtovtot in case $ngt overflows
	$ngtovtot = $ngt;
	$ovtotflag = 1;
      }
      else {
	$ovtot = $tot;
      }
    }
    # Create a new set of data to calculate the statistic of
    &shuffle(\@flat);		# Shuffle the list of values
    my $k = 0;
    for(my $i = 0; $i <= $#samples; $i++) {
      for(my $j = 0; $j < scalar(@{$samples[$i]}); $j++, $k++) {
	$samples[$i][$j] = $flat[$k];
      }
    }
    # Get the KW test statistic of this permutation, and see if it is greater
    # than that of the original data
    my($rT, $rdf) = &calckwstat(\@samples, \%ranks);
    if($rT > $$Tstat) {
      $ngt++;
      # Check for overflows in the counter of the number of permutations with
      # greater KW test statistic
      if($ngt == $ovngt) {	# We've overflowed -- bail out from the state
				# of affairs just before the total overflowed
	$$pvalue = $ngtovtot / $ovtot;
	return ($$pvalue < $siglevel ? 1 : 0);
      }
      else {
	$ovngt = $ngt;
      }
    }
    elsif($allgtflag) {		# pvalue is less than 1...
      $allgtflag = 0;
    }
    # Iterate
    for(my $i = $#acccount; $i >= 0; $i--) {
      $acccount[$i]--;		# When $acccount[0] gets decremented, we're
				# done -- the while condition will break.
      last if $acccount[$i] > 0;
				# Don't need to decrement the next counter
      $acccount[$i] = 10;
    }
  }
  # OK. If we get here, then we've looped through 10**accuracy randomly
  # generated permutations of the ranks. Generate the p-value.
  $$pvalue = ($allgtflag ? 1 : sprintf("0.%0*d", $accuracy, $ngt));
  return ($$pvalue < $siglevel ? 1 : 0);
}
  
# shuffle
# 
# Shuffle the data in the list passed by reference in the argument. This
# is done by taking stuff off either the head or the tail of the list and
# adding it to either the tail or the head of a temporary list, and then
# doing the same thing back again.

sub shuffle {
  my($list) = @_;

  my(@tmp);
  
  while(scalar(@$list) > 0) {
    if(rand() < 0.5) {
      if(rand() < 0.5) {
	push(@tmp, pop(@$list));
      }
      else {
	push(@tmp, shift(@$list));
      }
    }
    else {
      if(rand() < 0.5) {
	unshift(@tmp, pop(@$list));
      }
      else {
	unshift(@tmp, shift(@$list));
      }
    }
  }
  while(scalar(@tmp) > 0) {
    if(rand() < 0.5) {
      if(rand() < 0.5) {
	push(@$list, pop(@tmp));
      }
      else {
	push(@$list, shift(@tmp));
      }
    }
    else {
      if(rand() < 0.5) {
	unshift(@$list, pop(@tmp));
      }
      else {
	unshift(@$list, shift(@tmp));
      }
    }
  }
}

# rankdata
#
# Take an array of arrays of observations, and rank the data therein, also
# a sorted, flattened list of observations, and returning an array copy of the
# data.

sub rankdata {
  my($data, $ranks, $flat) = @_;

  my(@samples) = @$data;
  my(@copy);

  for(my $i = 0; $i <= $#samples; $i++) {
    my(@observations) = @{$samples[$i]};
    $copy[$i] = [ @observations ];
    push(@$flat, @observations);
  }
  @$flat = sort { $a <=> $b } @$flat;
  my $lastne = 0;
  for(my $i = 0; $i < scalar(@$flat); $i++) {
    if($$flat[$i] != $$flat[$i - 1]) {
      $$ranks{$$flat[$i - 1]} = $lastne + 1 + (($i - 1 - $lastne) / 2);
      $lastne = $i;
    }
  }
  $$ranks{$$flat[scalar(@$flat) - 1]} = $lastne + 1 + ((scalar(@$flat)
							- 1 - $lastne) / 2);
  return (@copy);
}

# calckwstat
#
# Return the Kruskal-Wallis test statistic given a set of data and its ranks.
# The "with-ties" version is used. See Sprent & Smeeton (2001, p. 201). Also
# return the number of degrees of freedom.

sub calckwstat {
  my($data, $ranks) = @_;

  my(@samples) = @$data;

  # We need the sum of squared ranks for each sample, ssr, and the mean
  # sum of squared ranks for each sample, ssk

  my($ssr, $ssk) = (0, 0);
  my $N = 0;
  my $k = 0;
  for(my $i = 0; $i <= $#samples; $i++) {
    my $ni = 0;
    my $si = 0;
    my @observations = @{$samples[$i]};
    for(my $j = 0; $j <= $#observations; $j++) {
#      print "$observations[$j] -> $$ranks{$observations[$j]}".
#	($j == $#observations ? "\n" : ", ");
      $si += $$ranks{$observations[$j]};
      $ssr += $$ranks{$observations[$j]} * $$ranks{$observations[$j]};
      $ni++;
      $N++;
    }
    $ssk += $si * $si / $ni;
    $k++;
  }

  # Now compute the test statistic, T.

  my $C = 0.25 * $N * ($N + 1) * ($N + 1);
  my $T = ($N - 1) * ($ssk - $C) / ($ssr - $C);

  return ($T, $k - 1);
}


# testchisq
#
# Return a boolean indicating whether or not the null hypothesis is rejected
# at the given level of significance for a ChiSq test.

sub testchisq {
  my($teststat, $degfree, $siglevel, $pvalue) = @_;

  $$pvalue = &gammq(($degfree / 2), ($teststat / 2));
  if($$pvalue < $siglevel) {
    return 1;
  }
  else {
    return 0;
  }
}

# gammq
#
# Return the incomplete gamma function Q(a,x) = 1 - P(a,x). This is taken
# from Numerical Recipes in C (2nd Edition) by Press et al (1992), Section 6.2,
# pp. 218-219. Translated into Perl, obviously...

sub gammq {
  my($a, $x) = @_;

  my($gamser, $gammcf, $gln);

  # Note that $gln is not used. Press et al give gln in case you want
  # ln(GAMMA(a)), and, though frankly I don't, I do want to give a faithful
  # rendition of the C in Perl.

  if($x < 0.0 || $a <= 0.0) {
    print STDERR "Error in gammq routine: ".($a <= 0.0 ? "a <= 0.0 ($a)" : "").
      ($x < 0.0 ? "x < 0.0 ($x) " : "")."\n";
    exit 1;
  }
  if($x < ($a + 1.0)) {
    # Use the series representation...
    &gser(\$gamser, $a, $x, \$gln);
    # ... and take its complement
    return (1.0 - $gamser);
  }
  else {
    # Use the continued fraction representation
    &gcf(\$gammcf, $a, $x, \$gln);
    return $gammcf;
  }
}

# gser
#
# Returns the incomplete gamma function P(a, x) evalutated by its series
# representation as gamser. Also returns ln(GAMMA(a)) as gln.

sub gser {
  my($gamser, $a, $x, $gln) = @_;

  my($n);
  my($sum, $del, $ap);

  # Maximum allowed number of iterations
  my $ITMAX = 200;
  # Relative accuracy
  my $EPS = 3.0e-14;

  $$gln = &gammln($a);

  if($x <= 0.0) {
    if($x < 0.0) {
      print STDERR "Error in gser routine: x < 0.0 ($x)\n";
      exit 1;
    }
    $$gamser = 0.0;
  }
  else {
    $ap = $a;
    $sum = 1.0 / $a;
    $del = $sum;
    for($n = 1; $n <= $ITMAX; $n++) {
      ++$ap;
      $del *= $x / $ap;
      $sum += $del;
      if(abs($del) < abs($sum) * $EPS) {
	$$gamser = $sum * exp(-$x + $a * log($x) - $$gln);
	return;
      }
    }
    print STDERR "a is too large ($a) and ITMAX too small ($ITMAX) in routine".
      " gser\n";
  }
}

# gcf
#
# Returns the incomplete gamma function Q(a, x) evaluated by its continued
# fraction representation as gammcf. Also returns ln(GAMMA(a)) as gln.

sub gcf {
  my($gammcf, $a, $x, $gln) = @_;

  my ($i);
  my ($an, $b, $c, $d, $del, $h);

  # Maximum allowed number of iterations
  my $ITMAX = 200;
  # Relative accuracy
  my $EPS = 3.0e-14;
  # Number near the smallest representable floating-point numebr
  my $FPMIN = 1.0e-60;

  $$gln = &gammln($a);
  # Set up for evaluating continued fraction by modified Lentz's method
  # (Section 5.2) with b_0 = 0
  $b = $x + 1.0 - $a;
  $c = 1.0 / $FPMIN;
  $d = 1.0 / $b;

  $h = $d;

  # Iterate to convergence
  for($i = 1; $i <= $ITMAX; $i++) {
    $an = -$i * ($i - $a);
    $b += 2.0;
    $d = ($an * $d) + $b;
    $d = $FPMIN if(abs($d) < $FPMIN);
    $c = $b + ($an / $c);
    $c = $FPMIN if(abs($c) < $FPMIN);
    $d = 1.0 / $d;
    $del = $d * $c;
    $h *= $del;
    last if(abs($del - 1.0) < $EPS);
  }
  if($i > $ITMAX) {
    print STDERR "a too large ($a) and ITMAX too small ($ITMAX) in gcf ".
      "routine\n";
    exit 1;
  }
  # Put factors in front
  $$gammcf = exp(-$x + ($a * log($x)) - $$gln) * $h;
}

# gammln
#
# Returns the value ln(GAMMA(xx)) for xx > 0

sub gammln {
  my($xx) = @_;
  my($x, $y, $tmp, $ser);

  my @cof = (76.18009172947146,
	     -86.50532032941677,
	     24.01409824083091,
	     -1.231739572450155,
	     0.1208650973866179e-2,
	     -0.5395239384953e-5);
  my $j;

  $y = $x = $xx;
  $tmp = $x + 5.5;
  $tmp -= ($x + 0.5) * log($tmp);
  # The next line and comment are written so as to check that the correct
  # number of zeros appear in the constant. (The second 1 appears above the
  # first + of j++ in the book.)
  $ser=1.000000000190015;
  #for (j=0;j<=5;j++) ser
  for($j = 0; $j <= 5; $j++) {
    $y++;
    $ser += $cof[$j] / $y;
  }
  return (-$tmp + log(2.5066282746310005 * $ser / $x));
}

# mychomp
#
# Remove end of line characters from a string

sub mychomp {
  my ($str) = @_;

  my $eol = substr($str, -1, 1);
  if($eol eq "\n") {
    $eol = substr($str, -2, 2);
    if($eol eq "\r\n") {
      return substr($str, 0, -2);
    }
    else {
      return substr($str, 0, -1);
    }
  }
  elsif($eol eq "\r") {
    return substr($str, 0, -1);
  }
  else {
    return $str;
  }
}

1;
